@@ -1,7 +1,9 @@
-__version__ = '3.0.1'
+__version__ = '3.5.0'
 
 from pymemcache.client.base import Client  # noqa
 from pymemcache.client.base import PooledClient  # noqa
+from pymemcache.client.hash import HashClient  # noqa
+from pymemcache.client.base import KeepaliveOpts  # noqa
 
 from pymemcache.exceptions import MemcacheError  # noqa
 from pymemcache.exceptions import MemcacheClientError  # noqa
