@@ -2,6 +2,8 @@
 
 from pymemcache.client.base import Client  # noqa
 from pymemcache.client.base import PooledClient  # noqa
+from pymemcache.client.hash import HashClient  # noqa
+from pymemcache.client.retrying import RetryingClient  # noqa
 
 from pymemcache.exceptions import MemcacheError  # noqa
 from pymemcache.exceptions import MemcacheClientError  # noqa
