@@ -12,6 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 import errno
+import platform
 import socket
 import six
 
@@ -37,7 +38,17 @@ VALID_STORE_RESULTS = {
     b'prepend': (b'STORED', b'NOT_STORED'),
     b'cas':     (b'STORED', b'EXISTS', b'NOT_FOUND'),
 }
-VALID_STRING_TYPES = (six.text_type, six.string_types)
+
+SOCKET_KEEPALIVE_SUPPORTED_SYSTEM = {
+    'Linux',
+}
+
+STORE_RESULTS_VALUE = {
+    b'STORED': True,
+    b'NOT_STORED': False,
+    b'NOT_FOUND':  None,
+    b'EXISTS': False
+}
 
 
 # Some of the values returned by the "stats" command
@@ -87,9 +98,9 @@ def check_key_helper(key, allow_unicode_keys, key_prefix=b''):
     if allow_unicode_keys:
         if isinstance(key, six.text_type):
             key = key.encode('utf8')
-    elif isinstance(key, VALID_STRING_TYPES):
+    elif isinstance(key, six.string_types):
         try:
-            if isinstance(key, bytes):
+            if isinstance(key, six.binary_type):
                 key = key.decode().encode('ascii')
             else:
                 key = key.encode('ascii')
@@ -102,7 +113,7 @@ def check_key_helper(key, allow_unicode_keys, key_prefix=b''):
     if len(key) > 250:
         raise MemcacheIllegalInputError("Key is too long: %r" % key)
     # second statement catches leading or trailing whitespace
-    elif len(parts) > 1 or parts[0] != key:
+    elif len(parts) > 1 or (parts and parts[0] != key):
         raise MemcacheIllegalInputError("Key contains whitespace: %r" % key)
     elif b'\00' in key:
         raise MemcacheIllegalInputError("Key contains null: %r" % key)
@@ -110,6 +121,61 @@ def check_key_helper(key, allow_unicode_keys, key_prefix=b''):
     return key
 
 
+def normalize_server_spec(server):
+    if isinstance(server, tuple) or server is None:
+        return server
+    if isinstance(server, list):
+        return tuple(server)  # Assume [host, port] provided.
+    if not isinstance(server, six.string_types):
+        raise ValueError('Unknown server provided: %r' % server)
+    if server.startswith('unix:'):
+        return server[5:]
+    if server.startswith('/'):
+        return server
+    if ':' not in server or server.endswith(']'):
+        host, port = server, 11211
+    else:
+        host, port = server.rsplit(':', 1)
+        port = int(port)
+    if host.startswith('['):
+        host = host.strip('[]')
+    return (host, port)
+
+
+class KeepaliveOpts(object):
+    """
+    A configuration structure to define the socket keepalive.
+
+    This structure must be passed to a client. The client will configure
+    its socket keepalive by using the elements of the structure.
+
+    Args:
+      idle: The time (in seconds) the connection needs to remain idle
+        before TCP starts sending keepalive probes. Should be a positive
+        integer most greater than zero.
+      intvl: The time (in seconds) between individual keepalive probes.
+        Should be a positive integer most greater than zero.
+      cnt: The maximum number of keepalive probes TCP should send before
+        dropping the connection. Should be a positive integer most greater
+        than zero.
+    """
+    __slots__ = ('idle', 'intvl', 'cnt')
+
+    def __init__(self, idle=1, intvl=1, cnt=5):
+        if idle < 1:
+            raise ValueError(
+                "The idle parameter must be greater or equal to 1.")
+        self.idle = idle
+        if intvl < 1:
+            raise ValueError(
+                "The intvl parameter must be greater or equal to 1.")
+        self.intvl = intvl
+        if cnt < 1:
+            raise ValueError(
+                "The cnt parameter must be greater or equal to 1.")
+        self.cnt = cnt
+
+
 class Client(object):
     """
     A client for a single memcached server.
@@ -216,10 +282,12 @@ class Client(object):
                  no_delay=False,
                  ignore_exc=False,
                  socket_module=socket,
+                 socket_keepalive=None,
                  key_prefix=b'',
                  default_noreply=True,
                  allow_unicode_keys=False,
-                 encoding='ascii'):
+                 encoding='ascii',
+                 tls_context=None):
         """
         Constructor.
 
@@ -241,6 +309,9 @@ class Client(object):
             misses. Defaults to False.
           socket_module: socket module to use, e.g. gevent.socket. Defaults to
             the standard library's socket module.
+          socket_keepalive: Activate the socket keepalive feature by passing
+            a KeepaliveOpts structure in this parameter. Disabled by default
+            (None). This feature is only supported on Linux platforms.
           key_prefix: Prefix of key. You can use this as namespace. Defaults
             to b''.
           default_noreply: bool, the default value for 'noreply' as passed to
@@ -253,22 +324,49 @@ class Client(object):
           The constructor does not make a connection to memcached. The first
           call to a method on the object will do that.
         """
-        self.server = server
+        self.server = normalize_server_spec(server)
         self.serde = serde or LegacyWrappingSerde(serializer, deserializer)
         self.connect_timeout = connect_timeout
         self.timeout = timeout
         self.no_delay = no_delay
         self.ignore_exc = ignore_exc
         self.socket_module = socket_module
+        self.socket_keepalive = socket_keepalive
+        user_system = platform.system()
+        if self.socket_keepalive is not None:
+            if user_system not in SOCKET_KEEPALIVE_SUPPORTED_SYSTEM:
+                raise SystemError(
+                    "Pymemcache's socket keepalive mechaniss doesn't "
+                    "support your system ({user_system}). If "
+                    "you see this message it mean that you tried to "
+                    "configure your socket keepalive on an unsupported "
+                    "system. To fix the problem pass `socket_"
+                    "keepalive=False` or use a supported system. "
+                    "Supported systems are: {systems}".format(
+                        user_system=user_system,
+                        systems=", ".join(sorted(
+                            SOCKET_KEEPALIVE_SUPPORTED_SYSTEM))
+                    )
+                )
+            if not isinstance(self.socket_keepalive, KeepaliveOpts):
+                raise ValueError(
+                    "Unsupported keepalive options. If you see this message "
+                    "it means that you passed an unsupported object within "
+                    "the param `socket_keepalive`. To fix it "
+                    "please instantiate and pass to socket_keepalive a "
+                    "KeepaliveOpts object. That's the only supported type "
+                    "of structure."
+                )
         self.sock = None
         if isinstance(key_prefix, six.text_type):
             key_prefix = key_prefix.encode('ascii')
-        if not isinstance(key_prefix, bytes):
+        if not isinstance(key_prefix, six.binary_type):
             raise TypeError("key_prefix should be bytes.")
         self.key_prefix = key_prefix
         self.default_noreply = default_noreply
         self.allow_unicode_keys = allow_unicode_keys
         self.encoding = encoding
+        self.tls_context = tls_context
 
     def check_key(self, key):
         """Checks key and add key_prefix."""
@@ -278,19 +376,49 @@ class Client(object):
     def _connect(self):
         self.close()
 
-        if isinstance(self.server, (list, tuple)):
-            sock = self.socket_module.socket(self.socket_module.AF_INET,
-                                             self.socket_module.SOCK_STREAM)
+        s = self.socket_module
+
+        if not isinstance(self.server, tuple):
+            sockaddr = self.server
+            sock = s.socket(s.AF_UNIX, s.SOCK_STREAM)
+
         else:
-            sock = self.socket_module.socket(self.socket_module.AF_UNIX,
-                                             self.socket_module.SOCK_STREAM)
+            sock = None
+            error = None
+            host, port = self.server
+            info = s.getaddrinfo(host, port, s.AF_UNSPEC, s.SOCK_STREAM,
+                                 s.IPPROTO_TCP)
+            for family, socktype, proto, _, sockaddr in info:
+                try:
+                    sock = s.socket(family, socktype, proto)
+                    if self.no_delay:
+                        sock.setsockopt(s.IPPROTO_TCP, s.TCP_NODELAY, 1)
+                    if self.tls_context:
+                        context = self.tls_context
+                        sock = context.wrap_socket(sock, server_hostname=host)
+                except Exception as e:
+                    error = e
+                    if sock is not None:
+                        sock.close()
+                        sock = None
+                else:
+                    break
+
+            if error is not None:
+                raise error
+
         try:
             sock.settimeout(self.connect_timeout)
-            sock.connect(self.server)
+            if self.socket_keepalive is not None:
+                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
+                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE,
+                                self.socket_keepalive.idle)
+                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL,
+                                self.socket_keepalive.intvl)
+                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT,
+                                self.socket_keepalive.cnt)
+            sock.connect(sockaddr)
             sock.settimeout(self.timeout)
-            if self.no_delay and sock.family == self.socket_module.AF_INET:
-                sock.setsockopt(self.socket_module.IPPROTO_TCP,
-                                self.socket_module.TCP_NODELAY, 1)
         except Exception:
             sock.close()
             raise
@@ -308,6 +436,8 @@ class Client(object):
             finally:
                 self.sock = None
 
+    disconnect_all = close
+
     def set(self, key, value, expire=0, noreply=None, flags=None):
         """
         The memcached "set" command.
@@ -471,6 +601,7 @@ class Client(object):
           the key didn't exist, False if it existed but had a different cas
           value and True if it existed and was changed.
         """
+        cas = self._check_cas(cas)
         return self._store_cmd(b'cas', {key: value}, expire, noreply,
                                flags=flags, cas=cas)[key]
 
@@ -628,7 +759,7 @@ class Client(object):
 
         Args:
           key: str, see class docs for details.
-          value: int, the amount by which to increment the value.
+          value: int, the amount by which to decrement the value.
           noreply: optional bool, False to wait for the reply (the default).
 
         Returns:
@@ -770,6 +901,32 @@ class Client(object):
         self._misc_cmd([cmd], b'quit', True)
         self.close()
 
+    def shutdown(self, graceful=False):
+        """
+        The memcached "shutdown" command.
+
+        This will request shutdown and eventual termination of the server,
+        optionally preceded by a graceful stop of memcached's internal state
+        machine. Note that the server needs to have been started with the
+        shutdown protocol command enabled with the --enable-shutdown flag.
+
+        Args:
+          graceful: optional bool, True to request a graceful shutdown with
+                    SIGUSR1 (defaults to False, i.e. SIGINT shutdown).
+        """
+        cmd = b'shutdown'
+        if graceful:
+            cmd += b' graceful'
+        cmd += b'\r\n'
+
+        # The shutdown command raises a server-side error if the shutdown
+        # protocol command is not enabled. Otherwise, a successful shutdown
+        # is expected to close the remote end of the transport.
+        try:
+            self._misc_cmd([cmd], b'shutdown', False)
+        except MemcacheUnexpectedCloseError:
+            pass
+
     def _raise_errors(self, line, name):
         if line.startswith(b'ERROR'):
             raise MemcacheUnknownCommandError(name)
@@ -784,7 +941,7 @@ class Client(object):
 
     def _check_integer(self, value, name):
         """Check that a value is an integer and encode it as a binary string"""
-        if not isinstance(value, six.integer_types):  # includes "long" on py2
+        if not isinstance(value, six.integer_types):
             raise MemcacheIllegalInputError(
                 '%s must be integer, got bad value: %r' % (name, value)
             )
@@ -798,7 +955,7 @@ class Client(object):
         The value will be (re)encoded so that we can accept strings or bytes.
         """
         # convert non-binary values to binary
-        if isinstance(cas, (six.integer_types, VALID_STRING_TYPES)):
+        if isinstance(cas, (six.integer_types, six.string_types)):
             try:
                 cas = six.text_type(cas).encode(self.encoding)
             except UnicodeEncodeError:
@@ -822,7 +979,7 @@ class Client(object):
         """
         This function is abstracted from _fetch_cmd to support different ways
         of value extraction. In order to use this feature, _extract_value needs
-        to be overriden in the subclass.
+        to be overridden in the subclass.
         """
         if expect_cas:
             _, key, flags, size, cas = line.split()
@@ -832,7 +989,12 @@ class Client(object):
             except Exception as e:
                 raise ValueError("Unable to parse line %s: %s" % (line, e))
 
-        buf, value = _readvalue(self.sock, buf, int(size))
+        value = None
+        try:
+            buf, value = _readvalue(self.sock, buf, int(size))
+        except MemcacheUnexpectedCloseError:
+            self.close()
+            raise
         key = remapped_keys[key]
         value = self.serde.deserialize(key, value, int(flags))
 
@@ -846,7 +1008,10 @@ class Client(object):
         remapped_keys = dict(zip(prefixed_keys, keys))
 
         # It is important for all keys to be listed in their original order.
-        cmd = name + b' ' + b' '.join(prefixed_keys) + b'\r\n'
+        cmd = name
+        if prefixed_keys:
+            cmd += b' ' + b' '.join(prefixed_keys)
+        cmd += b'\r\n'
 
         try:
             if self.sock is None:
@@ -855,9 +1020,14 @@ class Client(object):
             self.sock.sendall(cmd)
 
             buf = b''
+            line = None
             result = {}
             while True:
-                buf, line = _readline(self.sock, buf)
+                try:
+                    buf, line = _readline(self.sock, buf)
+                except MemcacheUnexpectedCloseError:
+                    self.close()
+                    raise
                 self._raise_errors(line, name)
                 if line == b'END' or line == b'OK':
                     return result
@@ -887,7 +1057,6 @@ class Client(object):
 
         extra = b''
         if cas is not None:
-            cas = self._check_cas(cas)
             extra += b' ' + cas
         if noreply:
             extra += b' noreply'
@@ -928,19 +1097,17 @@ class Client(object):
 
             results = {}
             buf = b''
+            line = None
             for key in keys:
-                buf, line = _readline(self.sock, buf)
+                try:
+                    buf, line = _readline(self.sock, buf)
+                except MemcacheUnexpectedCloseError:
+                    self.close()
+                    raise
                 self._raise_errors(line, name)
 
                 if line in VALID_STORE_RESULTS[name]:
-                    if line == b'STORED':
-                        results[key] = True
-                    if line == b'NOT_STORED':
-                        results[key] = False
-                    if line == b'NOT_FOUND':
-                        results[key] = None
-                    if line == b'EXISTS':
-                        results[key] = False
+                    results[key] = STORE_RESULTS_VALUE[line]
                 else:
                     raise MemcacheUnknownError(line[:32])
             return results
@@ -960,8 +1127,13 @@ class Client(object):
 
             results = []
             buf = b''
+            line = None
             for cmd in cmds:
-                buf, line = _readline(self.sock, buf)
+                try:
+                    buf, line = _readline(self.sock, buf)
+                except MemcacheUnexpectedCloseError:
+                    self.close()
+                    raise
                 self._raise_errors(line, cmd_name)
                 results.append(line)
             return results
@@ -990,8 +1162,11 @@ class PooledClient(object):
       max_pool_size: maximum pool size to use (going above this amount
                      triggers a runtime error), by default this is 2147483648L
                      when not provided (or none).
+      pool_idle_timeout: pooled connections are discarded if they have been
+                         unused for this many seconds. A value of 0 indicates
+                         that pooled connections are never discarded.
       lock_generator: a callback/type that takes no arguments that will
-                      be called to create a lock or sempahore that can
+                      be called to create a lock or semaphore that can
                       protect the pool from concurrent access (for example a
                       eventlet lock or semaphore could be used instead)
 
@@ -1001,6 +1176,9 @@ class PooledClient(object):
     in the pool. Your serde object must therefore be thread-safe.
     """
 
+    #: :class:`Client` class used to create new clients
+    client_class = Client
+
     def __init__(self,
                  server,
                  serde=None,
@@ -1011,32 +1189,38 @@ class PooledClient(object):
                  no_delay=False,
                  ignore_exc=False,
                  socket_module=socket,
+                 socket_keepalive=None,
                  key_prefix=b'',
                  max_pool_size=None,
+                 pool_idle_timeout=0,
                  lock_generator=None,
                  default_noreply=True,
                  allow_unicode_keys=False,
-                 encoding='ascii'):
-        self.server = server
+                 encoding='ascii',
+                 tls_context=None):
+        self.server = normalize_server_spec(server)
         self.serde = serde or LegacyWrappingSerde(serializer, deserializer)
         self.connect_timeout = connect_timeout
         self.timeout = timeout
         self.no_delay = no_delay
         self.ignore_exc = ignore_exc
         self.socket_module = socket_module
+        self.socket_keepalive = socket_keepalive
         self.default_noreply = default_noreply
         self.allow_unicode_keys = allow_unicode_keys
         if isinstance(key_prefix, six.text_type):
             key_prefix = key_prefix.encode('ascii')
-        if not isinstance(key_prefix, bytes):
+        if not isinstance(key_prefix, six.binary_type):
             raise TypeError("key_prefix should be bytes.")
         self.key_prefix = key_prefix
         self.client_pool = pool.ObjectPool(
             self._create_client,
             after_remove=lambda client: client.close(),
             max_size=max_pool_size,
+            idle_timeout=pool_idle_timeout,
             lock_generator=lock_generator)
         self.encoding = encoding
+        self.tls_context = tls_context
 
     def check_key(self, key):
         """Checks key and add key_prefix."""
@@ -1044,23 +1228,27 @@ class PooledClient(object):
                                 key_prefix=self.key_prefix)
 
     def _create_client(self):
-        client = Client(self.server,
-                        serde=self.serde,
-                        connect_timeout=self.connect_timeout,
-                        timeout=self.timeout,
-                        no_delay=self.no_delay,
-                        # We need to know when it fails *always* so that we
-                        # can remove/destroy it from the pool...
-                        ignore_exc=False,
-                        socket_module=self.socket_module,
-                        key_prefix=self.key_prefix,
-                        default_noreply=self.default_noreply,
-                        allow_unicode_keys=self.allow_unicode_keys)
-        return client
+        return self.client_class(
+            self.server,
+            serde=self.serde,
+            connect_timeout=self.connect_timeout,
+            timeout=self.timeout,
+            no_delay=self.no_delay,
+            # We need to know when it fails *always* so that we
+            # can remove/destroy it from the pool...
+            ignore_exc=False,
+            socket_module=self.socket_module,
+            socket_keepalive=self.socket_keepalive,
+            key_prefix=self.key_prefix,
+            default_noreply=self.default_noreply,
+            allow_unicode_keys=self.allow_unicode_keys,
+            tls_context=self.tls_context)
 
     def close(self):
         self.client_pool.clear()
 
+    disconnect_all = close
+
     def set(self, key, value, expire=0, noreply=None, flags=None):
         with self.client_pool.get_and_release(destroy_on_fail=True) as client:
             return client.set(key, value, expire=expire, noreply=noreply,
@@ -1068,9 +1256,8 @@ class PooledClient(object):
 
     def set_many(self, values, expire=0, noreply=None, flags=None):
         with self.client_pool.get_and_release(destroy_on_fail=True) as client:
-            failed = client.set_many(values, expire=expire, noreply=noreply,
-                                     flags=flags)
-            return failed
+            return client.set_many(values, expire=expire, noreply=noreply,
+                                   flags=flags)
 
     set_multi = set_many
 
@@ -1188,6 +1375,10 @@ class PooledClient(object):
             finally:
                 self.client_pool.destroy(client)
 
+    def shutdown(self, graceful=False):
+        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
+            client.shutdown(graceful)
+
     def __setitem__(self, key, value):
         self.set(key, value, noreply=True)
 
