@@ -21,15 +21,25 @@ import functools
 import json
 import os
 import mock
+import platform
+import re
 import socket
+import sys
 import unittest
 
 import pytest
 
-from pymemcache.client.base import PooledClient, Client
+from pymemcache.client.base import (
+    PooledClient,
+    Client,
+    normalize_server_spec,
+    KeepaliveOpts,
+    check_key_helper
+)
 from pymemcache.exceptions import (
     MemcacheClientError,
     MemcacheServerError,
+    MemcacheUnexpectedCloseError,
     MemcacheUnknownCommandError,
     MemcacheUnknownError,
     MemcacheIllegalInputError
@@ -39,6 +49,49 @@ from pymemcache import pool
 from pymemcache.test.utils import MockMemcacheClient
 
 
+# TODO: Use ipaddress module when dropping support for Python < 3.3
+def is_ipv6(address):
+    return re.match(r'^[0-9a-f:]+$', address)
+
+
+@pytest.mark.parametrize(
+    'key,allow_unicode_keys,key_prefix,ex_exception,ex_excinfo,ignore_py27',
+    [
+        (u'b'*251, True, b'',
+         MemcacheIllegalInputError, 'Key is too long', False),
+        (u'foo bar', True, b'',
+         MemcacheIllegalInputError, 'Key contains whitespace', False),
+        (u'\00', True, b'',
+         MemcacheIllegalInputError, 'Key contains null', False),
+        (None, True, b'', TypeError, None, False),
+        # The following test won't fail with a TypeError with python 2.7
+        (b"", False, '', TypeError, None, True),
+    ])
+@pytest.mark.unit()
+def test_check_key_helper_failing_conditions(key, allow_unicode_keys,
+                                             key_prefix, ex_exception,
+                                             ex_excinfo, ignore_py27):
+
+    if ignore_py27 and sys.version_info < (3, 0, 0):
+        pytest.skip("skipping for Python 2.7")
+    with pytest.raises(ex_exception) as excinfo:
+        check_key_helper(key, allow_unicode_keys, key_prefix)
+
+    # Allow to ignore the excinfo value. Different implementation (2.7,
+    # pypy, 3.x) comes with various error messages for the same thing.
+    # We just check the kind of the exception.
+    if ex_excinfo:
+        assert ex_excinfo in str(excinfo.value)
+
+
+@pytest.mark.unit()
+def test_check_key_helper():
+    assert check_key_helper(b"key", True, b'') == b"key"
+    assert check_key_helper("key", True) == b"key"
+    assert isinstance(check_key_helper(b"key", True), bytes)
+    assert check_key_helper("", True) == b""
+
+
 class MockSocket(object):
     def __init__(self, recv_bufs, connect_failure=None, close_failure=None):
         self.recv_bufs = collections.deque(recv_bufs)
@@ -52,7 +105,8 @@ class MockSocket(object):
 
     @property
     def family(self):
-        return socket.AF_INET
+        any_ipv6 = any(is_ipv6(c[0]) for c in self.connections)
+        return socket.AF_INET6 if any_ipv6 else socket.AF_INET
 
     def sendall(self, value):
         self.send_bufs.append(value)
@@ -103,7 +157,7 @@ class MockSocketModule(object):
         self.close_failure = close_failure
         self.sockets = []
 
-    def socket(self, family, type):
+    def socket(self, family, type, proto=0, fileno=None):
         socket = MockSocket(
             [],
             connect_failure=self.connect_failure,
@@ -111,6 +165,19 @@ class MockSocketModule(object):
         self.sockets.append(socket)
         return socket
 
+    def getaddrinfo(self, host, port, family=0, type=0, proto=0, flags=0):
+        family = family or (
+            socket.AF_INET6 if is_ipv6(host) else socket.AF_INET
+        )
+        type = type or socket.SOCK_STREAM
+        proto = proto or socket.IPPROTO_TCP
+        sockaddr = (
+            ('::1', 11211, 0, 0)
+            if family == socket.AF_INET6
+            else ('127.0.0.1', 11211)
+        )
+        return [(family, type, proto, '', sockaddr)]
+
     def __getattr__(self, name):
         return getattr(socket, name)
 
@@ -481,6 +548,9 @@ class TestClient(ClientTestMixin, unittest.TestCase):
 
     def test_cas_malformed(self):
         client = self.make_client([b'STORED\r\n'])
+        with pytest.raises(MemcacheIllegalInputError):
+            client.cas(b'key', b'value', None, noreply=False)
+
         with pytest.raises(MemcacheIllegalInputError):
             client.cas(b'key', b'value', 'nonintegerstring', noreply=False)
 
@@ -665,6 +735,19 @@ class TestClient(ClientTestMixin, unittest.TestCase):
         assert result is None
         assert client.sock is None
 
+    def test_shutdown(self):
+        client = self.make_client([MemcacheUnexpectedCloseError('shutdown')])
+        result = client.shutdown()
+        assert result is None
+
+    def test_shutdown_disabled(self):
+        def _shutdown():
+            client = self.make_client([b'ERROR: shutdown not enabled\r\n'])
+            client.shutdown()
+
+        with pytest.raises(MemcacheUnknownCommandError):
+            _shutdown()
+
     def test_replace_stored(self):
         client = self.make_client([b'STORED\r\n'])
         result = client.replace(b'key', b'value', noreply=False)
@@ -777,6 +860,20 @@ class TestClient(ClientTestMixin, unittest.TestCase):
         with pytest.raises(MemcacheServerError):
             _set()
 
+    def test_closing_socket_on_unexpected_closed_error(self):
+        client = self.make_client([
+            b'VALUE ',
+            MemcacheUnexpectedCloseError("foo bar"),
+        ])
+
+        def _set():
+            client.set(b'key', b'value', noreply=False)
+
+        with pytest.raises(MemcacheUnexpectedCloseError):
+            _set()
+
+        assert client.sock is None
+
     def test_set_unknown_error(self):
         client = self.make_client([b'foobarbaz\r\n'])
 
@@ -882,7 +979,7 @@ class TestClient(ClientTestMixin, unittest.TestCase):
         client = self.make_client([b'STAT fake_stats 1\r\n', b'END\r\n'])
         result = client.stats()
         assert client.sock.send_bufs == [
-            b'stats \r\n'
+            b'stats\r\n'
         ]
         assert result == {b'fake_stats': 1}
 
@@ -916,7 +1013,7 @@ class TestClient(ClientTestMixin, unittest.TestCase):
         ])
         result = client.stats()
         assert client.sock.send_bufs == [
-            b'stats \r\n'
+            b'stats\r\n'
         ]
         expected = {
             b'cmd_get': 2519,
@@ -1075,12 +1172,39 @@ class TestClient(ClientTestMixin, unittest.TestCase):
 
 @pytest.mark.unit()
 class TestClientSocketConnect(unittest.TestCase):
-    def test_socket_connect(self):
-        server = ("example.com", 11211)
+    def test_socket_connect_ipv4(self):
+        server = ('127.0.0.1', 11211)
 
         client = Client(server, socket_module=MockSocketModule())
         client._connect()
         assert client.sock.connections == [server]
+        assert client.sock.family == socket.AF_INET
+
+        timeout = 2
+        connect_timeout = 3
+        client = Client(
+            server, connect_timeout=connect_timeout, timeout=timeout,
+            socket_module=MockSocketModule())
+        client._connect()
+        assert client.sock.timeouts == [connect_timeout, timeout]
+
+        client = Client(server, socket_module=MockSocketModule())
+        client._connect()
+        assert client.sock.socket_options == []
+
+        client = Client(
+            server, socket_module=MockSocketModule(), no_delay=True)
+        client._connect()
+        assert client.sock.socket_options == [(socket.IPPROTO_TCP,
+                                               socket.TCP_NODELAY, 1)]
+
+    def test_socket_connect_ipv6(self):
+        server = ('::1', 11211)
+
+        client = Client(server, socket_module=MockSocketModule())
+        client._connect()
+        assert client.sock.connections == [server + (0, 0)]
+        assert client.sock.family == socket.AF_INET6
 
         timeout = 2
         connect_timeout = 3
@@ -1108,6 +1232,47 @@ class TestClientSocketConnect(unittest.TestCase):
             client._connect()
             assert client.sock.family == socket.AF_UNIX
 
+    @unittest.skipIf('Linux' != platform.system(),
+                     'Socket keepalive only support Linux platforms.')
+    def test_linux_socket_keepalive(self):
+        server = ('::1', 11211)
+        try:
+            client = Client(
+                server,
+                socket_module=MockSocketModule(),
+                socket_keepalive=KeepaliveOpts())
+            client._connect()
+        except SystemError:
+            self.fail("SystemError unexpectedly raised")
+        with self.assertRaises(ValueError):
+            # A KeepaliveOpts object is expected, a ValueError will be raised
+            Client(
+                server,
+                socket_module=MockSocketModule(),
+                socket_keepalive=True)
+
+    @mock.patch('platform.system')
+    def test_osx_socket_keepalive(self, platform_mock):
+        platform_mock.return_value = 'Darwin'
+        server = ('::1', 11211)
+        # For the moment the socket keepalive is only implemented for Linux
+        with self.assertRaises(SystemError):
+            Client(
+                server,
+                socket_module=MockSocketModule(),
+                socket_keepalive=KeepaliveOpts())
+
+    @mock.patch('platform.system')
+    def test_windows_socket_keepalive(self, platform_mock):
+        platform_mock.return_value = 'Windows'
+        server = ('::1', 11211)
+        # For the moment the socket keepalive is only implemented for Linux
+        with self.assertRaises(SystemError):
+            Client(
+                server,
+                socket_module=MockSocketModule(),
+                socket_keepalive=KeepaliveOpts())
+
     def test_socket_connect_closes_on_failure(self):
         server = ("example.com", 11211)
 
@@ -1213,6 +1378,45 @@ class TestPooledClient(ClientTestMixin, unittest.TestCase):
                                     [b'__FAKE_RESPONSE__\r\n'])
         self._default_noreply_true('flush_all', (), [b'__FAKE_RESPONSE__\r\n'])
 
+    def test_custom_client(self):
+        class MyClient(Client):
+            pass
+        client = PooledClient(('host', 11211))
+        client.client_class = MyClient
+        assert isinstance(client.client_pool.get(), MyClient)
+
+
+class TestPooledClientIdleTimeout(ClientTestMixin, unittest.TestCase):
+    def make_client(self, mock_socket_values, **kwargs):
+        mock_client = Client(None, **kwargs)
+        mock_client.sock = MockSocket(list(mock_socket_values))
+        client = PooledClient(None, pool_idle_timeout=60, **kwargs)
+        client.client_pool = pool.ObjectPool(lambda: mock_client)
+        return client
+
+    def test_free_idle(self):
+        class Counter(object):
+            count = 0
+
+            def increment(self, obj):
+                self.count += 1
+
+        removed = Counter()
+
+        client = self.make_client([b'VALUE key 0 5\r\nvalue\r\nEND\r\n']*2)
+        client.client_pool._after_remove = removed.increment
+        client.client_pool._idle_clock = lambda: 0
+
+        client.set(b'key', b'value')
+        assert removed.count == 0
+        client.get(b'key')
+        assert removed.count == 0
+
+        # Advance clock to beyond the idle timeout.
+        client.client_pool._idle_clock = lambda: 61
+        client.get(b'key')
+        assert removed.count == 1
+
 
 class TestMockClient(ClientTestMixin, unittest.TestCase):
     def make_client(self, mock_socket_values, **kwargs):
@@ -1323,3 +1527,46 @@ class TestRetryOnEINTR(unittest.TestCase):
             b'ue1\r\nEND\r\n',
         ])
         assert client[b'key1'] == b'value1'
+
+
+@pytest.mark.unit()
+class TestNormalizeServerSpec(unittest.TestCase):
+    def test_normalize_server_spec(self):
+        f = normalize_server_spec
+        assert f(None) is None
+        assert f(('127.0.0.1', 12345)) == ('127.0.0.1', 12345)
+        assert f(['127.0.0.1', 12345]) == ('127.0.0.1', 12345)
+        assert f('unix:/run/memcached/socket') == '/run/memcached/socket'
+        assert f('/run/memcached/socket') == '/run/memcached/socket'
+        assert f('localhost') == ('localhost', 11211)
+        assert f('localhost:12345') == ('localhost', 12345)
+        assert f('[::1]') == ('::1', 11211)
+        assert f('[::1]:12345') == ('::1', 12345)
+        assert f('127.0.0.1') == ('127.0.0.1', 11211)
+        assert f('127.0.0.1:12345') == ('127.0.0.1', 12345)
+
+        with pytest.raises(ValueError) as excinfo:
+            f({'host': 12345})
+        assert str(excinfo.value) == "Unknown server provided: {'host': 12345}"
+
+        with pytest.raises(ValueError) as excinfo:
+            f(12345)
+        assert str(excinfo.value) == "Unknown server provided: 12345"
+
+
+@pytest.mark.unit()
+class TestKeepaliveopts(unittest.TestCase):
+    def test_keepalive_opts(self):
+        kao = KeepaliveOpts()
+        assert (kao.idle == 1 and kao.intvl == 1 and kao.cnt == 5)
+        kao = KeepaliveOpts(idle=1, intvl=4, cnt=2)
+        assert (kao.idle == 1 and kao.intvl == 4 and kao.cnt == 2)
+        kao = KeepaliveOpts(idle=8)
+        assert (kao.idle == 8 and kao.intvl == 1 and kao.cnt == 5)
+        kao = KeepaliveOpts(cnt=8)
+        assert (kao.idle == 1 and kao.intvl == 1 and kao.cnt == 8)
+
+        with self.assertRaises(ValueError):
+            KeepaliveOpts(cnt=0)
+        with self.assertRaises(ValueError):
+            KeepaliveOpts(idle=-1)
